@@ -8,7 +8,7 @@ use std::ffi::OsStr;
 // use clipboard::{ClipboardContext, ClipboardProvider};
 use log::debug;
 use std::io::{stdin, stdout, Write};
-use std::path::{Path, PathBuf};
+use std::path::PathBuf;
 use std::sync::{Arc, Mutex};
 use std::thread;
 use std::time::Duration;
@@ -19,6 +19,7 @@ use termion::input::TermRead;
 use termion::raw::IntoRawMode;
 use termion::{clear, cursor, screen};
 
+//frequency to detect terminal size change
 const DETECTION_INTERVAL: u64 = 500;
 
 pub fn run(arg: PathBuf) -> Result<(), MyError> {
@@ -104,6 +105,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
         let input = stdin.next();
         let mut state = state_run.lock().unwrap();
         let mut screen = screen_run.lock().unwrap();
+        screen.flush()?;
         let mut nums = nums_run.lock().unwrap();
         let len = state.list.len();
         let y = state.layout.y;
@@ -211,6 +213,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                 print!("{}", screen::ToAlternateScreen);
                                 if let Err(e) = state.open_file(nums.index) {
                                     print_warning(e, y);
+                                    screen.flush()?;
                                     continue;
                                 }
                                 print!("{}", screen::ToAlternateScreen);
@@ -224,6 +227,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                 Some(true_path) => match std::fs::File::open(true_path) {
                                     Err(e) => {
                                         print_warning(e, y);
+                                        screen.flush()?;
                                         continue;
                                     }
                                     Ok(_) => {
@@ -248,6 +252,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                             std::env::set_current_dir(&state.current_dir)
                                         {
                                             print_warning(e, y);
+                                            screen.flush()?;
                                             continue;
                                         }
                                         state.update_list()?;
@@ -261,6 +266,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                     print!("{}", screen::ToAlternateScreen);
                                     if let Err(e) = state.open_file(nums.index) {
                                         print_warning(e, y);
+                                        screen.flush()?;
                                         continue;
                                     }
                                     print!("{}", screen::ToAlternateScreen);
@@ -274,6 +280,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                 match std::fs::File::open(&item.file_path) {
                                     Err(e) => {
                                         print_warning(e, y);
+                                        screen.flush()?;
                                         continue;
                                     }
                                     Ok(_) => {
@@ -299,6 +306,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                             std::env::set_current_dir(&state.current_dir)
                                         {
                                             print_warning(e, y);
+                                            screen.flush()?;
                                             continue;
                                         }
                                         state.update_list()?;
@@ -366,6 +374,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                     }
                                     if let Err(e) = std::env::set_current_dir(&state.current_dir) {
                                         print_warning(e, y);
+                                        screen.flush()?;
                                         continue;
                                     }
                                     state.update_list()?;
@@ -379,6 +388,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                     state.current_dir = parent_p.to_path_buf();
                                     if let Err(e) = std::env::set_current_dir(&state.current_dir) {
                                         print_warning(e, y);
+                                        screen.flush()?;
                                         continue;
                                     }
                                     state.update_list()?;
@@ -396,18 +406,21 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
 
                                             if nums.index
                                                 >= (state.layout.terminal_row
-                                                    - (STARTING_POINT + 3))
+                                                    - (STARTING_POINT + 1))
                                                     .into()
                                             {
-                                                nums.skip = (nums.index - 3) as u16;
+                                                nums.skip = (nums.index - 1) as u16;
                                                 clear_and_show(&state.current_dir);
                                                 state.list_up(nums.skip);
-                                                state.move_cursor(&nums, STARTING_POINT + 3);
+                                                state.move_cursor(&nums, STARTING_POINT + 1);
                                             } else {
                                                 nums.skip = 0;
                                                 clear_and_show(&state.current_dir);
                                                 state.list_up(0);
-                                                state.move_cursor(&nums, (nums.index + 3) as u16);
+                                                state.move_cursor(
+                                                    &nums,
+                                                    (nums.index as u16) + STARTING_POINT,
+                                                );
                                             }
                                         }
                                         None => {
@@ -426,6 +439,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                     }
                 }
 
+                //select mode
                 Key::Char('V') => {
                     if len == 0 {
                         continue;
@@ -446,20 +460,20 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                         if let Some(Ok(key)) = input {
                             match key {
                                 Key::Char('j') | Key::Down => {
-                                    if nums.index == len - 1 {
+                                    if len == 0 || nums.index == len - 1 {
                                         continue;
-                                    } else if y == state.layout.terminal_row - 4
+                                    } else if y >= state.layout.terminal_row - 4
                                         && len
                                             > (state.layout.terminal_row - STARTING_POINT) as usize
                                                 - 1
                                     {
-                                        nums.inc_skip();
                                         nums.go_down();
+                                        nums.inc_skip();
 
                                         if nums.index > start_pos {
                                             let mut item = state.list.get_mut(nums.index).unwrap();
                                             item.selected = true;
-                                        } else if nums.index < start_pos {
+                                        } else {
                                             let mut item =
                                                 state.list.get_mut(nums.index - 1).unwrap();
                                             item.selected = false;
@@ -484,30 +498,28 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                         clear_and_show(&state.current_dir);
                                         state.list_up(nums.skip);
                                         state.move_cursor(&nums, y + 1);
-                                        screen.flush()?;
                                     }
                                 }
 
                                 Key::Char('k') | Key::Up => {
-                                    if y == STARTING_POINT {
+                                    if nums.index == 0 {
                                         continue;
-                                    } else if y == STARTING_POINT + 3 && nums.skip != 0 {
-                                        nums.dec_skip();
+                                    } else if y <= STARTING_POINT + 3 && nums.skip != 0 {
                                         nums.go_up();
+                                        nums.dec_skip();
 
                                         if nums.index >= start_pos {
                                             let mut item =
                                                 state.list.get_mut(nums.index + 1).unwrap();
                                             item.selected = false;
-                                        } else if nums.index < start_pos {
+                                        } else {
                                             let mut item = state.list.get_mut(nums.index).unwrap();
                                             item.selected = true;
                                         }
 
                                         clear_and_show(&state.current_dir);
                                         state.list_up(nums.skip);
-                                        state.move_cursor(&nums, STARTING_POINT + 3);
-                                        screen.flush()?;
+                                        state.move_cursor(&nums, y);
                                     } else {
                                         nums.go_up();
 
@@ -523,7 +535,6 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                         clear_and_show(&state.current_dir);
                                         state.list_up(nums.skip);
                                         state.move_cursor(&nums, y - 1);
-                                        screen.flush().unwrap();
                                     }
                                 }
 
@@ -597,47 +608,31 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                     screen.flush()?;
 
                                     state.registered.clear();
-                                    let mut count: usize = 0;
                                     let clone = state.list.clone();
-                                    let iter = clone.iter().filter(|item| item.selected);
-                                    let total_selected = iter.clone().count();
-                                    for (i, item) in iter.enumerate() {
-                                        print!(
-                                            " {}{}{}",
-                                            cursor::Goto(2, 2),
-                                            clear::CurrentLine,
-                                            display_count(i, total_selected)
-                                        );
-                                        match item.file_type {
-                                            FileType::Directory => {
-                                                if let Err(e) =
-                                                    state.remove_and_yank_dir(item.clone())
-                                                {
-                                                    print_warning(e, y);
-                                                    break;
-                                                }
-                                            }
-                                            FileType::File | FileType::Symlink => {
-                                                if let Err(e) =
-                                                    state.remove_and_yank_file(item.clone())
-                                                {
-                                                    print_warning(e, y);
-                                                    break;
-                                                }
-                                            }
-                                        }
-                                        count += 1;
+                                    let selected: Vec<ItemInfo> =
+                                        clone.into_iter().filter(|item| item.selected).collect();
+                                    let total = selected.len();
+
+                                    if let Err(e) = state.remove_and_yank(&selected, true) {
+                                        print_warning(e, y);
+                                        screen.flush()?;
+                                        break;
                                     }
+
                                     clear_and_show(&state.current_dir);
                                     state.update_list()?;
+                                    let new_len = state.list.len();
+                                    if usize::from(nums.skip) >= new_len {
+                                        nums.reset();
+                                    }
                                     state.list_up(nums.skip);
 
                                     let duration = duration_to_string(start.elapsed());
                                     let delete_message: String = {
-                                        if count == 1 {
+                                        if total == 1 {
                                             format!("1 item deleted [{}]", duration)
                                         } else {
-                                            let mut count = count.to_string();
+                                            let mut count = total.to_string();
                                             count.push_str(&format!(
                                                 " items deleted [{}]",
                                                 duration
@@ -648,12 +643,14 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                     print_info(delete_message, y);
                                     print!(" ");
 
-                                    let new_len = state.list.len();
                                     if new_len == 0 {
                                         nums.reset();
                                         state.move_cursor(&nums, STARTING_POINT);
                                     } else if nums.index > new_len - 1 {
-                                        let new_y = y - (nums.index - (new_len - 1)) as u16;
+                                        let mut new_y = y - (nums.index - (new_len - 1)) as u16;
+                                        if new_y < 3 {
+                                            new_y = 3;
+                                        }
                                         nums.index = new_len - 1;
                                         state.move_cursor(&nums, new_y)
                                     } else {
@@ -694,6 +691,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                     }
                 }
 
+                //toggle sortkey
                 Key::Char('t') => {
                     match state.sort_by {
                         SortKey::Name => {
@@ -710,6 +708,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                     state.move_cursor(&nums, STARTING_POINT);
                 }
 
+                //delete
                 Key::Char('d') => {
                     if len == 0 {
                         continue;
@@ -729,29 +728,13 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                         let start = Instant::now();
                                         screen.flush()?;
 
-                                        state.registered.clear();
-                                        let item = state.get_item(nums.index)?.clone();
-                                        match item.file_type {
-                                            FileType::Directory => {
-                                                print!(
-                                                    " {}{}1/1",
-                                                    cursor::Goto(2, 2),
-                                                    clear::CurrentLine
-                                                );
-                                                if let Err(e) = state.remove_and_yank_dir(item) {
-                                                    print_warning(e, y);
-                                                    state.move_cursor(&nums, y);
-                                                    break 'delete;
-                                                }
-                                            }
-                                            FileType::File | FileType::Symlink => {
-                                                if let Err(e) = state.remove_and_yank_file(item) {
-                                                    clear_and_show(&state.current_dir);
-                                                    print_warning(e, y);
-                                                    state.move_cursor(&nums, y);
-                                                    break 'delete;
-                                                }
-                                            }
+                                        let target = state.get_item(nums.index)?.clone();
+                                        let target = vec![target];
+
+                                        if let Err(e) = state.remove_and_yank(&target, true) {
+                                            print_warning(e, y);
+                                            screen.flush()?;
+                                            continue;
                                         }
 
                                         clear_and_show(&state.current_dir);
@@ -786,6 +769,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                     }
                 }
 
+                //yank
                 Key::Char('y') => {
                     if len == 0 {
                         continue;
@@ -821,6 +805,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                     }
                 }
 
+                //put
                 Key::Char('p') => {
                     if state.registered.is_empty() {
                         continue;
@@ -829,8 +814,10 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                     let start = Instant::now();
                     screen.flush()?;
 
-                    if let Err(e) = state.put_items() {
+                    let targets = state.registered.clone();
+                    if let Err(e) = state.put_items(&targets, None) {
                         print_warning(e, y);
+                        screen.flush()?;
                         continue;
                     }
 
@@ -850,12 +837,13 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                     state.move_cursor(&nums, y);
                 }
 
+                //rename
                 Key::Char('c') => {
                     if len == 0 {
                         continue;
                     }
                     print!("{}", cursor::Show);
-                    let item = state.get_item(nums.index).unwrap();
+                    let item = state.get_item(nums.index).unwrap().clone();
 
                     let mut rename = item.file_name.chars().collect::<Vec<char>>();
                     print!(
@@ -878,19 +866,24 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                     let rename = rename.iter().collect::<String>();
                                     let mut to = state.current_dir.clone();
                                     to.push(rename);
-                                    if let Err(e) =
-                                        std::fs::rename(Path::new(&item.file_path), Path::new(&to))
-                                    {
+                                    if let Err(e) = std::fs::rename(&item.file_path, &to) {
                                         print!("{}", cursor::Hide);
                                         print_warning(e, y);
                                         break;
                                     }
 
+                                    state.manipulations.manip_list.push(ManipKind::Rename(
+                                        RenamedFile {
+                                            original_name: item.file_path.clone(),
+                                            new_name: to,
+                                        },
+                                    ));
+                                    state.manipulations.count = 0;
+
+                                    print!("{}", cursor::Hide);
                                     clear_and_show(&state.current_dir);
                                     state.update_list()?;
                                     state.list_up(nums.skip);
-
-                                    print!("{}", cursor::Hide);
                                     state.move_cursor(&nums, y);
                                     break;
                                 }
@@ -956,6 +949,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                     }
                 }
 
+                //filter mode
                 Key::Char('/') => {
                     if len == 0 {
                         continue;
@@ -1079,6 +1073,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                     print!("{}", cursor::Hide);
                 }
 
+                //shell mode
                 Key::Char(':') => {
                     print!(" {}{}:", cursor::Goto(2, 2), clear::CurrentLine,);
                     print!("{}", cursor::Show);
@@ -1213,6 +1208,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                             }
                                         } else {
                                             print_warning("zoxide not installed?", y);
+                                            break 'command;
                                         }
                                     }
 
@@ -1233,6 +1229,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                                         ) {
                                                             print!("{}", cursor::Hide);
                                                             print_warning(e, y);
+                                                            screen.flush()?;
                                                             continue 'main;
                                                         }
                                                         if let Err(e) =
@@ -1240,6 +1237,7 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                                                         {
                                                             print!("{}", cursor::Hide);
                                                             print_warning(e, y);
+                                                            screen.flush()?;
                                                             continue 'main;
                                                         }
                                                         break 'empty;
@@ -1353,6 +1351,156 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
                     }
                 }
 
+                //undo
+                Key::Char('u') => {
+                    let mani_len = state.manipulations.manip_list.len();
+                    if mani_len < state.manipulations.count + 1 {
+                        continue;
+                    }
+                    if let Some(manipulation) = state
+                        .manipulations
+                        .manip_list
+                        .get(mani_len - state.manipulations.count - 1)
+                    {
+                        match manipulation.clone() {
+                            ManipKind::Rename(m) => {
+                                if let Err(e) = std::fs::rename(&m.new_name, &m.original_name) {
+                                    print_warning(e, y);
+                                    screen.flush()?;
+                                    continue;
+                                }
+                                state.manipulations.count += 1;
+                                clear_and_show(&state.current_dir);
+                                state.update_list()?;
+                                state.list_up(nums.skip);
+                                print_info("Undone [rename]", y);
+                            }
+                            ManipKind::Put(m) => {
+                                for x in m.put {
+                                    //todo: should not use remove_file actually
+                                    if let Err(e) = std::fs::remove_file(&x) {
+                                        print_warning(e, y);
+                                        screen.flush()?;
+                                        continue;
+                                    }
+                                }
+                                state.manipulations.count += 1;
+                                clear_and_show(&state.current_dir);
+                                state.update_list()?;
+                                state.list_up(nums.skip);
+                                print_info("Undone [put]", y);
+                            }
+                            ManipKind::Delete(m) => {
+                                let targets = trash_to_info(&state.trash_dir, m.trash)?;
+                                if let Err(e) = state.put_items(&targets, Some(m.dir)) {
+                                    print_warning(e, y);
+                                    screen.flush()?;
+                                    continue;
+                                }
+                                state.manipulations.count += 1;
+                                clear_and_show(&state.current_dir);
+                                state.update_list()?;
+                                state.list_up(nums.skip);
+                                print_info("Undone [delete]", y);
+                            }
+                        }
+
+                        let new_len = state.list.len();
+                        if new_len == 0 {
+                            nums.reset();
+                            state.move_cursor(&nums, STARTING_POINT);
+                        } else if nums.index > new_len - 1 {
+                            let new_y = y - (nums.index - (new_len - 1)) as u16;
+                            nums.index = new_len - 1;
+                            state.move_cursor(&nums, new_y)
+                        } else {
+                            state.move_cursor(&nums, y);
+                        }
+                        screen.flush()?;
+                    }
+                }
+
+                //redo
+                Key::Ctrl('r') => {
+                    let mani_len = state.manipulations.manip_list.len();
+                    if mani_len == 0
+                        || state.manipulations.count == 0
+                        || mani_len < state.manipulations.count
+                    {
+                        continue;
+                    }
+                    if let Some(manipulation) = state
+                        .manipulations
+                        .manip_list
+                        .get(mani_len - state.manipulations.count)
+                    {
+                        let manipulation = manipulation.clone();
+                        match manipulation {
+                            ManipKind::Rename(m) => {
+                                if let Err(e) = std::fs::rename(&m.original_name, &m.new_name) {
+                                    print_warning(e, y);
+                                    screen.flush()?;
+                                    continue;
+                                }
+                                state.manipulations.count -= 1;
+                                clear_and_show(&state.current_dir);
+                                state.update_list()?;
+                                state.list_up(nums.skip);
+                                print_info("Redone [rename]", y);
+                            }
+                            ManipKind::Put(m) => {
+                                if let Err(e) = state.put_items(&m.original, Some(m.dir.clone())) {
+                                    print_warning(e, y);
+                                    screen.flush()?;
+                                    continue;
+                                }
+                                state.manipulations.count -= 1;
+                                clear_and_show(&state.current_dir);
+                                state.update_list()?;
+                                state.list_up(nums.skip);
+                                print_info("Redone [put]", y);
+                            }
+                            ManipKind::Delete(m) => {
+                                if let Err(e) = state.remove_and_yank(&m.original, false) {
+                                    print_warning(e, y);
+                                    screen.flush()?;
+                                    continue;
+                                }
+                                state.manipulations.count -= 1;
+                                clear_and_show(&state.current_dir);
+                                state.update_list()?;
+                                state.list_up(nums.skip);
+                                print_info("Redone [delete]", y);
+                            }
+                        }
+
+                        let new_len = state.list.len();
+                        if new_len == 0 {
+                            nums.reset();
+                            state.move_cursor(&nums, STARTING_POINT);
+                        } else if nums.index > new_len - 1 {
+                            let new_y = y - (nums.index - (new_len - 1)) as u16;
+                            nums.index = new_len - 1;
+                            state.move_cursor(&nums, new_y)
+                        } else {
+                            state.move_cursor(&nums, y);
+                        }
+                        screen.flush()?;
+                    }
+                }
+
+                //debug print for undo/redo
+                // Key::Char('P') => {
+                //     print_info(
+                //         format!(
+                //             "{:?} count: {}",
+                //             state.manipulations, state.manipulations.count
+                //         ),
+                //         y,
+                //     );
+                // }
+
+                //exit by ZZ
                 Key::Char('Z') => {
                     print!(" {}{}Z", cursor::Goto(2, 2), clear::CurrentLine,);
                     print!("{}", cursor::Show);
@@ -1408,15 +1556,17 @@ pub fn run(arg: PathBuf) -> Result<(), MyError> {
         screen.flush()?;
     }
 
-    //When exits, restore the cursor
-    print!("{}", cursor::Restore);
     let state = state_run.lock().unwrap();
-    //If layout was refreshed, go back to main screen to restore the previous state
-    state.write_session(session_file_path)?;
     let mut screen = screen_run.lock().unwrap();
+
+    //Save session, restore screen state and cursor
+    state.write_session(session_file_path)?;
     write!(screen, "{}", screen::ToMainScreen)?;
-    screen.suspend_raw_mode()?;
+    write!(screen, "{}", cursor::Restore)?;
     screen.flush()?;
+
+    //Back to normal mode
+    screen.suspend_raw_mode()?;
     Ok(())
 }
 
